"""
DSLIBRARY - run your data science code anywhere.

"""
__version__ = "0.0.1"

# environment variables
from .front import ENV_DSLIBRARY_TARGET, ENV_DSLIBRARY_SPEC, ENV_DSLIBRARY_TOKEN
# main base class
from .front import DSLibrary
# run models
from .run_models import ModelRunner

# if a package called 'mmlibrary' is installed it becomes the default target
try:
    import mmlibrary
except ImportError:
    mmlibrary = None


def instance():
    """
    Get an instance of mmlibrary methods.  Configuration is through environment variables.

    To run locally in the current folder:   (this is the default)
      DSLIBRARY_TARGET=local

    To run locally with a specific default folder:
      DSLIBRARY_TARGET=local:<folder>

    To run against a REST API:
      DSLIBRARY_TARGET=https://hostname:port/path/to/api/
      DSLIBRARY_TOKEN=<access token, or credentials>

    To use a custom implementation:
      DSLIBRARY_TARGET=package.ClassName:value:name=value

    Parameter values and data input/output locations/formats and such are defined by setting this variable to
    JSON (see ModelRunner).
      DSLIBRARY_SPEC={...}

    :returns:  An instance of the MMLibrary() base class.
    """
    from .to_local import DSLibraryLocal
    from .transport.to_rest import DSLibraryViaREST
    from .transport.to_volume import DSLibraryViaVolume
    from .transport.to_mmlibrary import DSLibraryViaMMLibrary

    import importlib
    import json
    import os
    spec = json.loads(os.environ.get(ENV_DSLIBRARY_SPEC) or "{}")
    target = os.environ.get(ENV_DSLIBRARY_TARGET) or ("mmlibrary" if mmlibrary else "local")
    target_parts = target.split(":")
    # point to a REST API
    if target_parts[0] in ("http", "https"):
        token = os.environ.get(ENV_DSLIBRARY_TOKEN)
        return DSLibraryViaREST(url=target, token=token)
    # gather arguments
    args = []
    kwargs = {}
    for arg in target_parts[1:]:
        arg_parts = arg.split('=', maxsplit=1)
        if len(arg_parts) == 1:
            args.append(arg)
        else:
            kwargs[arg_parts[0]] = arg_parts[1]
    if spec:
        kwargs["spec"] = spec
    # a strictly local instance
    if target_parts[0] == "local":
        return DSLibraryLocal(*args, **kwargs)
    # delegation to 'mmlibrary'
    if target_parts[0] == "mmlibrary":
        return DSLibraryViaMMLibrary(*args, **kwargs)
    # writes through a shared volume, i.e. so that a sidecar can perform communications for us
    if target_parts[0] == "volume":
        return DSLibraryViaVolume(*args, **kwargs)
    # we fall through to custom implementation support
    cls_parts = target_parts[0].split(".")
    try:
        module = importlib.import_module('.'.join(cls_parts[:-1]))
        cls = getattr(module, cls_parts[-1])
        return cls(*args, **kwargs)
    except ImportError:
        raise ValueError(f"Unrecognized {ENV_DSLIBRARY_TARGET} implementation: {target_parts[0]}")



"""
The 'other side' of communications for MMLibraryViaVolume.  Watches for requests appearing in a folder as JSON
files, and transmits them to a remote REST API.
"""
import os
import json
import time

from ..to_rest import DSLibraryViaREST


class VolumeWatcher(object):
    def __init__(self, volume: str, mmlibrary_rest: DSLibraryViaREST):
        self._volume = volume
        self._mm = mmlibrary_rest

    def scan_forever(self, callback=None, interval: float=0.05):
        """
        Scan repeatedly.  The supplied callback can raise an exception to stop the process.
        """
        while True:
            time.sleep(interval)
            if callback:
                callback()
            self.scan_once()

    def scan_once(self):
        """
        Scan for new requests.
        """
        for f in os.listdir(self._volume):
            if not f.endswith(".json"):
                continue
            try:
                fn = os.path.join(self._volume, f)
                with open(fn) as f_r:
                    request = json.load(f_r)
            except ValueError:
                continue
            response = self.process(request)
            # send response (if request file still exists)
            if os.path.exists(fn):
                response_fn = fn + ".response"
                response_tmp = fn + ".tmp"
                # write to temporary file
                with open(response_tmp, 'wb') as f_w:
                    if isinstance(response, (bytes, bytearray)):
                        f_w.write(response)
                    elif isinstance(response, str):
                        f_w.write(response.encode("utf-8"))
                    else:
                        json.dump(response, f_w)
                # rename to expected location
                os.rename(response_tmp, response_fn)

    def process(self, request: dict):
        return self._mm._do_comm(
            method=request["method"], path=request["path"], params=request["params"],
            data=request["data"], as_json=request["as_json"]
        )





def connect_to_database(path: str, library: str=None, for_write: bool=False, **kwargs):
    """
    Try, by various means, to connect to a database.

    :param path:        URI of database.
    :param library:     A specific library to use, if available
    :param for_write:   Whether to enable write operations.
    """
    # determine flavor of connection and work out which libraries to try
    libraries = []
    if library:
        libraries.append(library)
    path_parts = path.split(":")
    if path_parts[0] == "jdbc":
        path_parts = path_parts[1:]
    if path_parts[0] in ("postgres", "postgresql"):
        libraries.append("psycopg2")
    if path_parts[0] == "mysql":
        libraries.append("pymysql")
    libraries.append("JayDeBeApi")
    for lib in libraries:
        method = f"_connect_{lib}"
        if method in locals():
            conn = locals()[method](path=path, **kwargs)
            if conn:
                return conn


def _connect_psycopg2(path, **kwargs):
    try:
        import psycopg2
    except ImportError:
        return
    dsn = ""
    # TODO split up path to get host, port, database, also need username, password
    return psycopg2.connect(dsn)


# TODO pymysql
# TODO JayDeBeApi
# TODO sqlite3, but we have to resolve the file's location



"""
Run a model.
"""
import json
import os
import typing
import subprocess
import sys

from pexpect import split_command_line

from dslibrary import ENV_DSLIBRARY_TOKEN
from dslibrary.front import DSLibraryException, ENV_DSLIBRARY_TARGET, ENV_DSLIBRARY_SPEC, METRICS_ALIAS
from dslibrary.metadata import Metadata


class ModelRunner(object):
    """
    Each instance sets up for one run of one model.
    """
    def __init__(self, uri: str=None, user: str=None, run_id: str=None, entry_point: str = "main", mlflow: dict=None):
        """
        The constructor provides some context which is optional when directly invoking local code, and required
        when invoking code remotely.

        :param uri:             Unique identifier for the model.
        :param user:            Which user is running the model.
        :param run_id:          Run ID needs to be specified when running in MLFlow.
        :param entry_point:     Which entry point is being called.
        :param mlflow:          MLFlow-related settings.
        """
        self.uri = uri or ""
        self.user = user or ""
        self.run_id = run_id or ""
        self.entry_point = entry_point
        # TODO document & validate these
        self.mlflow = mlflow or {}
        self.inputs = {}
        self.outputs = {}
        self.parameters = {}

    def get_metadata(self):
        """
        Obtain metadata for the selected model.
        """
        # TODO code me ... depends quite a bit on where the model is and how it is packaged, not possible without more context

    def _to_absolute(self, uri: str):
        """
        Convert local paths to absolute paths except when they match a certain pattern.
        """
        # TODO test this (and make sure it seems reasonable to use '~')
        if ":" in uri:
            # no changes to URIs
            return uri
        if uri.startswith("~/"):
            # specifically point to files within the sandbox
            return "./" + uri[1:]
        if uri.startswith("/"):
            # already absolute
            return uri
        # convert to absolute path
        return os.path.abspath(uri)

    def set_input(self, name: str, uri: str, **kwargs):
        """
        Specify where data will come from for a particular named input.
        :param name:    Name of input.
        :param uri:     A path to a local file, or the URI of remote data.  Or, a URI specifying a sql or nosql data
                        source.
        :param kwargs:  Additional parameters to support the various data sources, following fsspec for file-like
                        sources.
        """
        self.inputs[name] = {"uri": self._to_absolute(uri), **kwargs}
        return self

    def set_output(self, name: str, uri: str, **kwargs):
        """
        Specify where data should go for a particular named output.

        A format can be chosen by specifying "format".
          csv or tab -- remaining arguments are sent to pandas.to_csv()
          json - remaining arguments are sent to pandas.to_json()
          etc.

        :param name:    Name of output.
        :param uri:     See set_input()
        :param kwargs:  See set_input()
        """
        self.outputs[name] = {"uri": self._to_absolute(uri), **kwargs}
        return self

    def set_parameter(self, name: str, value):
        """
        Specify a value for one of the parameters.
        """
        self.parameters[name] = value
        return self

    def set_parameters(self, params: dict):
        """
        Specify values for multiple parameters.
        """
        self.parameters.update(params)
        return self

    def send_metrics_to(self, uri: str=None, mlflow: bool=None, **kwargs):
        """
        Determine where metrics will be stored (for read and write).  Parameters are equivalent to set_output(), in
        that metrics can be sent to a particular location.

        :param uri:         A URI or filename.
        :param mlflow:      Specify True to have the model log its metrics through MLFlow.
        :param kwargs:      Additional arguments.
        """
        if mlflow is not None:
            self.mlflow["metrics"] = bool(mlflow)
        # TODO shouldn't the input and the output be separate?
        self.inputs[METRICS_ALIAS] = {"uri": uri, **kwargs}
        self.outputs[METRICS_ALIAS] = {"uri": uri, **kwargs}
        return self

    def _generate_spec(self):
        spec = {
            "uri": self.uri,
            "user": self.user,
            "run_id": self.run_id,
            "entry_point": self.entry_point,
            "inputs": self.inputs,
            "outputs": self.outputs,
            "parameters": self.parameters
        }
        if self.mlflow:
            spec["mlflow"] = self.mlflow
        return spec

    def run_method(self, method: typing.Callable, path: str=None):
        """
        As long as the target 'method' does not uses subprocesses or threads, this method can be used to invoke it.
        """
        # store specification
        orig_env = {ENV_DSLIBRARY_TARGET: os.environ.get(ENV_DSLIBRARY_TARGET), ENV_DSLIBRARY_SPEC: os.environ.get(ENV_DSLIBRARY_SPEC)}
        os.environ[ENV_DSLIBRARY_TARGET] = f"local:{path or '.'}"
        os.environ[ENV_DSLIBRARY_SPEC] = json.dumps(self._generate_spec())
        # the method makes calls to dslibrary, following rules set above
        method()
        # restore environment
        for k, v in orig_env.items():
            os.environ[k] = v or ""

    def run_local(self, path: str, extra_env: dict=None):
        """
        Execute a python, R or notebook based model in a subprocess.
        """
        # verify target exists, split apart path
        if not os.path.exists(path):
            raise DSLibraryException(f"Path not found: {path}")
        folder, fn = os.path.split(path)
        f_base, f_ext = os.path.splitext(fn)
        # if only a path is specified we can look up 'entry_point' and run its official command
        cmd = None
        if os.path.isdir(path):
            folder, fn = path, ""
            meta = Metadata.from_folder(path)
            entry_point = self.entry_point or "main"
            if entry_point not in meta.entry_points:
                raise DSLibraryException(f"Entry point {entry_point} not found in model at {path}")
            cmd_str = meta.entry_points[entry_point].command
            if not cmd_str:
                raise DSLibraryException(f"Entry point {entry_point} for {path} does not define a command")
            cmd = split_command_line(cmd_str)
        # if we're not using metadata we have to infer how to execute the code
        if not cmd:
            cmd = self.infer_command_from_path(path)
        # work out environment variables
        env = {
            ENV_DSLIBRARY_TARGET: f"local:{folder}",
            ENV_DSLIBRARY_SPEC: json.dumps(self._generate_spec()),
            **(extra_env or {})
        }
        # call the model
        result = subprocess.run(cmd, cwd=folder, env=env, check=True)
        # TODO report failures in a more elegant way

    @staticmethod
    def infer_command_from_path(path: str):
        """
        When we only know the name of a source file we have to work out how to execute it.
        """
        _, fn = os.path.split(path)
        _, f_ext = os.path.splitext(fn)
        f_ext = f_ext.lower()
        # if we're not using metadata we have to infer how to execute the code
        if f_ext == ".py":
            cmd = [sys.executable, path]
        elif f_ext == ".r":
            cmd = ["RScript", path]
        elif f_ext == ".ipynb":
            cmd = ["nbconvert", "--to", "notebook", "--execute", "--inplace", path]
        else:
            raise DSLibraryException(f"unrecognized executable extension: {f_ext} in {fn}")
        return cmd

    def prepare_remote_run(self, local_rest_url: str, rest_access_token: str=None) -> dict:
        """
        Generically prepare for a remote run, assuming the remote will use make REST calls to access its data.  Returns
        the environment variables that must be set when the remote executes.  These can go in a k8s job spec, for
        instance.
        :param local_rest_url:  URL of REST service.
        :param rest_access_token: Access token to secure communications.
        :return:  Environment variables to set, as a {}.
        """
        env = {}
        if local_rest_url:
            env[ENV_DSLIBRARY_TARGET] = local_rest_url
        if rest_access_token:
            env[ENV_DSLIBRARY_TOKEN] = rest_access_token
        env[ENV_DSLIBRARY_SPEC] = json.dumps(self._generate_spec())
        return env

    def run_mlflow(self, model_uri: str):
        """
        Use mlflow.run() to execute an MLFlow model.

        The trick will be how to set the environment variables in the remote model such that the interception takes
        place and sends data back to us.  Could initially only be supported when running locally with an environment
        variable.
        """
        # TODO code me



"""
Adapter to "mmlibrary".

The "mmlibrary" package can be installed locally to provide a default custom target.
"""
import io
import json

from dslibrary.front import DSLibraryException
from dslibrary.metadata import Metadata
from dslibrary.utils.file_utils import write_stream_with_read_on_close
from dslibrary import DSLibrary

try:
    import mmlibrary
except ImportError:
    mmlibrary = None


class DSLibraryViaMMLibrary(DSLibrary):
    def __init__(self, *args, _mm=None, **kwargs):
        """
        :param _mm:    You can reference a custom set of functions instead of a package.
        """
        self._mm = _mm or mmlibrary
        super(DSLibraryViaMMLibrary, self).__init__(*args, **kwargs)

    def get_metadata(self):
        return Metadata()

    def get_parameters(self):
        if not self._mm:
            return {}
        if hasattr(self._mm, "get_parameters"):
            return self._mm.get_parameters()
        if hasattr(self._mm, "param_dictionary"):
            return self._mm.param_dictionary
        return {}

    def get_parameter(self, parameter_name: str, default=None):
        if not self._mm:
            return default
        try:
            if hasattr(self._mm, "get_parameter"):
                return self._mm.get_parameter(parameter_name)
            if hasattr(self._mm, "get_argument"):
                return self._mm.get_argument(parameter_name)
        except KeyError:
            return default

    def _opener(self, path: str, mode: str, **kwargs) -> io.IOBase:
        if 'w' in mode:
            def finalize(fh):
                self._mm.save_binary_to_resource(path, fh.read())
            return write_stream_with_read_on_close(w_mode=mode, r_mode='rb', on_close=finalize)
        try:
            if mode == 'rb':
                return io.BytesIO(self._mm.get_binary_from_resource(path))
            if mode == 'r':
                return io.StringIO(self._mm.get_binary_from_resource(path).decode('utf-8'))
        except ValueError:
            raise FileNotFoundError(f"not found: {path}")
        raise DSLibraryException(f"Unsupported mode: {mode}")

    def open_run_data(self, filename: str, mode: str='rb') -> io.RawIOBase:
        if "r" in mode:
            data = json.loads(self._mm.get_temporary_data() or b'{}')
            if filename not in data:
                raise FileNotFoundError(f"not found: {filename}")
            if mode == 'r':
                return io.StringIO(data[filename])
            return io.BytesIO((data[filename]).encode("utf-8"))
        def finalize(fh):
            data = json.loads(self._mm.get_temporary_data() or b'{}')
            if "a" in mode:
                if filename not in data:
                    data[filename] = ""
                data[filename] += fh.read()
            else:
                data[filename] = fh.read()
            self._mm.save_temporary_data(json.dumps(data).encode('utf-8'))
        return write_stream_with_read_on_close(w_mode=mode, r_mode='r', on_close=finalize)

    def get_next_scoring_request(self, timeout: float=None) -> (dict, None):
        raise DSLibraryException("not implemented in self._mm, use get_parameter() for all fields of a single scoring request")

    def send_score(self, score):
        self._mm.report_score(score)

    def get_sql_connection(self, resource_name: str, for_write: bool=False, database: str=None, **kwargs):
        return self._mm.get_db_connection(resource_name)
